
import collections
import inspect
import re

import numpy as np
from numpy.random import randint


# Alphabet for dichotomous random variables:
F2 = (0, 1)


def get_arity(func):
    """Get arity of a function (number of arguments)."""
    return len(inspect.getargspec(func).args)


class SimpleCCA(collections.Sequence):

    # Construction:

    def __init__(self, rule, state):
        self.rule = rule
        self.state = state
        self.arity = get_arity(rule)

    @classmethod
    def create_random(cls, rule, size, alphabet=F2):
        state = [alphabet[i] for i in randint(len(alphabet), size=size)]
        return cls(rule, state)

    # Sequence protocol:

    def __getitem__(self, index):
        return self.state[index]

    def __setitem__(self, index, value):
        self.state[index] = value

    def __len__(self):
        return len(self.state)

    def __iter__(self):
        return iter(self.state)

    # General protocols:

    def __str__(self):
        return ''.join(map(str, self.state))

    __repr__ = __str__

    # CCA methods:

    def step(self):
        rule = self.rule
        state = self.state
        arity = self.arity
        new_size = len(self) - arity + 1
        new_state = [rule(*state[i:i+arity])
                     for i in range(new_size)]
        return SimpleCCA(rule, new_state)


def is_even(num):
    """Check if number is even (divisible by two)."""
    return num % 2 == 0


class Layouter:

    """
    Compute minimal layout of a CCA containing a set of variables (CellRef)
    and provide tools to address and extract them from a simulation.

        A0  A1  A2  A3  A4
          B0  B1  B2  B3
            C1  C2  C3
              D1  D2
                E1
    """

    def __init__(self, rule, alphabet_size, output_rvs):

        self.rule = rule
        self.num_outputs = len(output_rvs)
        self.alphabet_size = alphabet_size

        arity = get_arity(rule)

        # Since the update rule for each cell depends its neighbor cells, the
        # number of well-defined cells (and hence the size of the underlying
        # array) decreases at each time step (no boundary conditions):
        self.shrink = arity - 1
        # The picture above shows how this is supposed to take place. In
        # particular, the x-index of a cell with the same column number
        # decreases by [shrink] every two time steps. At odd time steps it
        # decreases by [floor(shrink/2)]. The exact rule can be found in the
        # real_index() method:
        self._left = np.array([self.shrink, self.shrink//2])
        self._right = np.array([self.shrink, (self.shrink+1)//2])

        # Compute required size and offsets to make all indices non-negative:
        min_x = min(self.left_ancestor(t, x) for t, x in output_rvs)
        max_x = max(self.right_ancestor(t, x) for t, x in output_rvs)
        # don't want numpy.int64 (conflicts with sympy):
        self.size = int(max_x - min_x + 1)
        self.xoffs = -int(min_x)

        self.input_shape = (alphabet_size,) * self.size
        self.output_shape = (alphabet_size,) * self.num_outputs

        # group output variables by time, so they can easily be looked up
        # later on:
        self.output_at = collections.defaultdict(list)
        for rv in output_rvs:
            self.output_at[rv.t].append(rv)
        self.max_t = max(self.output_at)

    def left_ancestor(self, t, x):
        """Get the root at the left backward light cone boundary."""
        return x - self._left.dot(divmod(t, 2))

    def right_ancestor(self, t, x):
        """Get the root at the right backward light cone boundary."""
        return x + self._right.dot(divmod(t, 2))

    def real_index(self, t, x):
        """Translate a cell position to the real index used internally."""
        return self.left_ancestor(t, x) + self.xoffs

    def __call__(self, state):
        """
        Simulate a CCA given an input state and extract the values of the
        initially specified output variables (given as CellRef) in the
        process.
        """
        cca = SimpleCCA(self.rule, state)
        for t in range(self.max_t + 1):
            if t > 0:
                cca = cca.step()
            for rv in self.output_at.get(t, ()):
                yield (rv, cca[self.real_index(t, rv.x)])

    def evolution_picture(self, state):
        cca = SimpleCCA(self.rule, state)
        for t in range(self.max_t + 1):
            if t > 0:
                cca = cca.step()
            line = ' '.join(map(str, cca.state))
            fill = ' ' * (self.shrink * t)
            yield fill + line + fill

    def vars_picture(self):
        for t in range(self.max_t + 1):
            size = self.size - self.shrink * t
            cca = ['--'] * size
            for rv in self.output_at.get(t, ()):
                cca[self.real_index(t, rv.x)] = rv.name
            line = '  '.join(map(str, cca))
            fill = ' ' * (self.shrink * t * 2)
            yield fill + line + fill

    def dual_picture(self, state):
        evol_pict = self.evolution_picture(state)
        vars_pict = self.vars_picture()
        for a, b in zip(evol_pict, vars_pict):
            yield a + ' | ' + b

    def random_input_state(self):
        return randint(self.alphabet_size, size=self.size)


def col2num(s):
    """Convert excel-like column string to 0-based index."""
    num = 0
    for c in s:
        num = num * 26 + ord(c) - ord('A') + 1
    return num - 1


def num2col(n):
    """Convert 0-based index to excel-like column string."""
    n += 1
    s = ''
    while n:
        n, r = divmod(n - 1, 26)
        s = chr(ord('A') + r) + s
    return s


class CellRef:

    def __init__(self, t, x):
        self.t = t
        self.x = x

    def __iter__(self):
        yield self.t
        yield self.x

    @property
    def name(self):
        return chr(self.t + ord('A')) + str(self.x)

    @classmethod
    def parse(cls, expr):
        m = re.match(r'^([A-Z]+)(-?\d+)$', expr.strip().upper())
        if not m:
            raise ValueError('Invalid cell name: {!r}'.format(expr))
        t = m.group(1)
        x = m.group(2)
        return cls(col2num(t), int(x))

    @classmethod
    def parse_set(cls, expr):
        return set(map(cls.parse, expr.split(',')))

    def __repr__(self):
        return num2col(self.t) + '_' + self.x

    def __hash__(self):
        return hash((self.t, self.x))

    def __eq__(self, other):
        return other == (self.t, self.x)


class Quantity:

    def __init__(self, expr, symbol, has_conditional, args):
        self.expr = expr
        self.symbol = symbol
        self.has_conditional = has_conditional
        self.args = args

    def __call__(self, ev, index):
        name = self.symbol
        if self.has_conditional:
            name += '_c'
        elif name == 'I' and len(self.args) == 3:
            name = 'I3'
        func = getattr(ev, name)
        args = ({index(rv) for rv in arg}
                for arg in self.args)
        return func(*args)

    @classmethod
    def from_string(cls, expr):
        m = re.match(r'^([HI])\((.+)\)$', expr.strip())
        if not m:
            raise ValueError('Invalid format: {!r}'.format(expr))
        func_str = m.group(1)
        args_str = m.group(2)
        if func_str == 'I':
            args = args_str.split(':')
        else:
            args = [args_str]
        has_conditional = '|' in args[-1]
        if has_conditional:
            args[-1:] = args[-1].split('|', 1)
        args = [CellRef.parse_set(arg) for arg in args]
        return cls(expr, func_str, has_conditional, args)
