
from functools import partial, reduce, wraps
import itertools

import numpy as np



class ProbabilityDistribution:

    def __init__(self, data, log):
        self.data = data
        self.log = log

    def h(self, p):
        """Compute one term in the entropy sum."""
        return - p * self.log(p) if p else 0

    def H(self):
        """Compute entropy from a numpy array of probabilities."""
        flags = ('refs_ok',)
        return sum(self.h(p.item()) for p in np.nditer(self.data, flags))

    def marginal(self, remove):
        """Return a marginal probability table."""
        return ProbabilityDistribution(self.data.sum(tuple(remove)),
                                       self.log)


def get_bits(num):
    """Return tuple of indices corresponding to 1-bits."""
    return tuple(i for i in range(num.bit_length())
                if num & (1 << i))


def bit_or(a, b):
    return a | b


def mkset(a):
    return a if isinstance(a, set) else {a}


def set_func(func):
    @wraps(func)
    def wrapper(self, *args):
        return func(self, *(mkset(a) for a in args))
    return wrapper


def fmt(s, *args):
    return s.format(*(','.join(map(str, arg)) for arg in args))


class EntropyVector:

    """
    Vector of entropies of all subsets of variables in a probability
    distribution.
    """

    def __init__(self, vec):
        self._vec = vec

    @classmethod
    def from_ptab(cls, ptab):
        """Create from a probability distribution."""
        dims = len(ptab.data.shape)
        size = 2 ** dims
        return cls([ptab.marginal(get_bits(size - 1 - i)).H()
                    for i in range(size)])

    @set_func
    def H(self, a):
        """Compute entropy H(a)."""
        bits = (1<<i for i in a)
        index = reduce(bit_or, bits, 0)
        return self._vec[index]

    @set_func
    def H_c(self, a, b):
        """Compute conditional entropy H(a|b)."""
        return self.H(a | b) - self.H(b)

    @set_func
    def I(self, a, b):
        """Compute mutual information I(a:b)."""
        return self.H(a) + self.H(b) - self.H(a | b)

    @set_func
    def I_c(self, a, b, c):
        """Compute conditional mutual information I(a:b|c)."""
        return self.H_c(a, c) + self.H_c(b, c) - self.H_c(a | b, c)

    @set_func
    def I3(self, a, b, c):
        """Compute mutual information I(a:b:c)."""
        return self.I(a, b) - self.I_c(a, b, c)

    H.fmt = partial(fmt, 'H({})')
    I.fmt = partial(fmt, 'I({}:{})')
    H_c.fmt = partial(fmt, 'H({}|{})')
    I_c.fmt = partial(fmt, 'I({}:{}|{})')
    I3.fmt = partial(fmt, 'I({}:{}:{})')

    def _itm(self, func, *args):
        return (func.fmt(*args), func(*args))
