# /usr/bin/env python3
# encoding: utf-8
"""
Usage:
    sim_cca [-r RULE] [-a SIZE] [-s COUNT | -p | -n] QUANTITIES...

Options:
    -a SIZE, --alphabet SIZE    Alphabet size           [default:         2]
    -r RULE, --rule RULE        Local update rule       [default:  a,b: a^b]
    -s COUNT, --samples COUNT   For monte-carlo mode
    -p, --picture               Draw a picture
    -n, --numeric               Numeric calculations

Examples:
    sim_cca -r 'a,b: abs(a-b)' -a5 'I(A0:B0:B1)' 'I(B0:B1|A0)'
    sim_cca -r 'a,b,c: a and (b or c)' -a2 'I(A0,B0:B1)' 'H(B0)' -p
"""

import docopt


from .it import EntropyVector, ProbabilityDistribution
from .cca import Layouter, Quantity
from . import  numeric_math


def analyze_output(quantity_strings):
    quantities = list(map(Quantity.from_string, quantity_strings))
    output_rvs = {rv for q in quantities for arg in q.args for rv in arg}
    output_rvs = list(output_rvs)
    # if ordered by time, the output order will be left invariant:
    output_rvs.sort(key=lambda rv: (rv.t, rv.x))
    return quantities, output_rvs


def monte_carlo(layouter, samples):
    """Probability estimation using monte-carlo method."""
    ptab = numeric_math.zeros(layouter.output_shape)
    for k in range(samples):
        results = layouter(layouter.random_input_state())
        index = tuple(val for rv, val in results)
        ptab[index] += 1
    ptab /= samples
    return ProbabilityDistribution(ptab, numeric_math.log)


def uniform_input_distribution(layouter, math):
    return math.full(layouter.input_shape,
                     math.rational(1, layouter.alphabet_size**layouter.size))


def pdist_based(layouter, math):
    """Simple numeric probability summation."""
    # independent, uniformly ditributed input layer:
    input_ptab = uniform_input_distribution(layouter, math)
    output_ptab = math.zeros(layouter.output_shape)
    for state, prob in math.ndenumerate(input_ptab):
        results = layouter(state)
        index = tuple(val for rv, val in results)
        output_ptab[index] += prob
    return ProbabilityDistribution(output_ptab, math.log)


def main(args=None):

    opts = docopt.docopt(__doc__, args)

    alphabet_size = int(opts['--alphabet'])
    rule = eval('lambda ' + opts['--rule'])
    picture = opts['--picture']
    samples = int(opts['--samples'] or '0')
    numeric = opts['--numeric']

    quantities, output_rvs = analyze_output(opts['QUANTITIES'])
    layouter = Layouter(rule, alphabet_size, output_rvs)

    if picture:
        print("\n".join(layouter.vars_picture()))
        return

    elif samples:
        ptab = monte_carlo(layouter, samples)

    else:
        if numeric:
            from . import numeric_math as math
        else:
            from . import symbolic_math as math
        ptab = pdist_based(layouter, math)

    ev = EntropyVector.from_ptab(ptab)
    results = {q.expr: q(ev, output_rvs.index)
               for q in quantities}

    for k, v in results.items():
        print(k, '=', math.simplify(v))


if __name__ == '__main__':
    main()
